/*
    Drools5 Integration Helper
    Copyright (C) 2009  Mathieu Boretti mathieu.boretti@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.boretti.drools.integration.drools5.implementation;

import java.lang.annotation.Annotation;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.boretti.drools.integration.drools5.DroolsInterface;
import org.boretti.drools.integration.drools5.DroolsServiceType;
import org.boretti.drools.integration.drools5.annotations.DroolsService;
import org.boretti.drools.integration.drools5.exceptions.NotAnInterfaceDroolsError;
import org.boretti.drools.integration.drools5.exceptions.NotAnnotatedClassDroolsError;
import org.boretti.drools.integration.drools5.exceptions.NotAnnotatedInterfaceDroolsError;
import org.boretti.drools.integration.drools5.implementation.DroolsInterfaceImplementation;
import org.boretti.drools.integration.drools5.implementation.DroolsCompiledInterfaceImplementation;
import org.drools.RuleBase;
import org.drools.StatelessSession;

public class DroolsProviderImplementation {
	
	private final Logger logger = Logger.getLogger(this.getClass());
	
	private final Map<String,DroolsInterfaceImplementation<?>> cache = new HashMap<String,DroolsInterfaceImplementation<?>>();
	
	private DroolsProviderImplementation() {}
	
	private static final DroolsProviderImplementation instance = new DroolsProviderImplementation();
	
	public static final DroolsProviderImplementation getInstance() {
		return instance;
	}
	
	@SuppressWarnings("unchecked")
	public <T> T getService(Class<T> clazz) {
		if (logger.isDebugEnabled()) logger.debug("Gettting service for "+clazz);
		validNullability(clazz,"Invalid interface - can't be null");
		if (!clazz.isInterface()) {
			String msg = "Not an interface "+clazz;
			logger.error(msg);
			throw new NotAnInterfaceDroolsError(msg);
		}
		DroolsService ds = getAnnotation(clazz);
		if(ds==null) {
			String msg = "Invalid interface (missing annotation) "+clazz;
			logger.error(msg);
			throw new NotAnnotatedInterfaceDroolsError(msg);
		}
		DroolsServiceType dst=ds.type();
		String resource = ds.resourceName();
		validNullability(dst,"Invalid type parameter of annotation - can't be null");
		DroolsInterfaceImplementation<T> inter=getDroolsImplementation(dst,resource,clazz);
		if (logger.isDebugEnabled()) logger.debug("Creating proxy using "+clazz+" from "+dst+" using resource "+resource);
		return (T) Proxy.newProxyInstance(clazz.getClassLoader(),new Class[] {clazz,DroolsInterface.class},new DroolsProxy<T>(clazz,inter));
	}
	
	/**
	 * 
	 * @param clazz
	 * @return
	 * @since 1.0.0
	 */
	public <T> RuleBase getRuleBase(Class<T> clazz) {
		if (logger.isDebugEnabled()) logger.debug("Gettting service for "+clazz);
		validNullability(clazz,"Invalid class or interface - can't be null");
		DroolsService ds = getAnnotation(clazz);
		if(ds==null) {
			if (clazz.isInterface()) {
				String msg = "Invalid interface (missing annotation) "+clazz;
				logger.error(msg);
				throw new NotAnnotatedInterfaceDroolsError(msg);
			} else {
				String msg = "Invalid class (missing annotation) "+clazz;
				logger.error(msg);
				throw new NotAnnotatedClassDroolsError(msg);
			}
		}
		DroolsServiceType dst=ds.type();
		String resource = ds.resourceName();
		validNullability(dst,"Invalid type parameter of annotation - can't be null");
		return getDroolsImplementation(dst,resource,clazz).getRuleBase();
	}
	
	/**
	 * 
	 * @param clazz
	 * @param type
	 * @param resourceName
	 * @return
	 * @since 1.1.0
	 */
	@SuppressWarnings("unchecked")
	public <T> RuleBase getRuleBase(Class<T> clazz,DroolsServiceType type,String resourceName) {
		validNullability(clazz,"Class can't be null");
		validNullability(type,"Type can't be null");
		validNullability(resourceName,"ResourceName can't be null");
		String key = clazz.getName()+":"+type+":"+resourceName;
		synchronized(cache) {
			DroolsInterfaceImplementation<?> element = cache.get(key);
			if (element!=null) {
				return element.getRuleBase();
			}
			element = getDroolsImplementation(type,resourceName,clazz);
			cache.put(key,element);
			return element.getRuleBase();
		}
	}
	
	private <T> DroolsInterfaceImplementation<T> getDroolsImplementation(DroolsServiceType dst,String resource,Class<T> clazz) {
		if (DroolsServiceType.COMPILED.equals(dst)) {
			if (resource==null || resource.trim().equalsIgnoreCase("")) resource = clazz.getName()+".cdrl";
			return new DroolsCompiledInterfaceImplementation<T>(clazz, resource);
		} else if (DroolsServiceType.SOURCE.equals(dst)){
			if (resource==null || resource.trim().equalsIgnoreCase("")) resource = clazz.getName()+".drl";
			return new DroolsSourceInterfaceImplementation<T>(clazz,resource);
		} else {
			if (resource==null || resource.trim().equalsIgnoreCase("")) resource = clazz.getName()+".xml";
			return new DroolsXMLInterfaceImplementation<T>(clazz,resource);
		}
	}
	
	private void validNullability(Object o,String msg) {
		if (o==null) {
			logger.error(msg);
			throw new IllegalArgumentException(msg);
		}
	}
	
	private <T> DroolsService getAnnotation(Class<T> clazz) {
		Annotation as[] = clazz.getAnnotations();
		for(Annotation a:as) {
			if (a.annotationType().equals(DroolsService.class)) return (DroolsService)a;
		}
		return null;
	}
	
	/**
	 * 
	 * @param clazz
	 * @param type
	 * @param resourceName
	 * @param error
	 * @param reference
	 * @param args
	 * @since 1.1.0
	 */
	@SuppressWarnings("unchecked")
	public <T> void runCondition(Class<T> clazz,
			DroolsServiceType type,
			String resourceName,
			Class<? extends Error> error,
			Object reference,
			Object args[]) {
		validNullability(error,"Error can't be null");
		validNullability(reference,"Reference can't be null");
		validNullability(args,"Args can't be null");
		if (logger.isDebugEnabled()) logger.debug("Doing PreCondition check for "+clazz+":"+type+":"+resourceName+":"+error);
		RuleBase rb = getRuleBase(clazz,type,resourceName);
		Logger l = Logger.getLogger(clazz);
		StatelessSession ss = rb.newStatelessSession();
		new DroolsWorkingMemoryLogger(ss);
		ss.setGlobal(FIELD_NAME_LOGGER, l);
		List<Object> al = new ArrayList<Object>();
		al.add(reference);
		for(Object o:args) al.add(o);
		Iterator<Object> i = ss.executeWithResults(al).iterateObjects();
		while(i.hasNext()) {
			Object o = i.next();
			if (error.isAssignableFrom(o.getClass())) {
				Error e = error.cast(o);
				logger.error("Error thrown from pre-condition : "+e,e);
				throw e;
			}
		}
	}
	
	/**
	 * 
	 * @param clazz
	 * @param type
	 * @param resourceName
	 * @param error
	 * @param reference
	 * @param args
	 * @since 1.1.0
	 */
	public <T> void runPreCondition(Class<T> clazz,
			DroolsServiceType type,
			String resourceName,
			Class<? extends Error> error,
			Object reference,
			Object args[]) {
		runCondition(clazz, type, resourceName, error, reference, args);
	}
	
	/**
	 * 
	 * @param clazz
	 * @param type
	 * @param resourceName
	 * @param error
	 * @param reference
	 * @param args
	 * @since 1.1.0
	 */
	public <T> void runPostCondition(Class<T> clazz,
			DroolsServiceType type,
			String resourceName,
			Class<? extends Error> error,
			Object reference,
			Object args[]) {
		runCondition(clazz, type, resourceName, error, reference, args);
	}
	
	/**
	 * @since 1.1.0
	 */
	private static final String FIELD_NAME_LOGGER = "droolsLogger";

}
